/*
 * bulkdata.c - Bulkdata core
 *
 * Copyright (C) 2022-2024, IOPSYS Software Solutions AB.
 *
 * Author: Amin Ben Romdhane <amin.benromdhane@iopsys.eu>
 * Author: Omar Kallel <omar.kallel@pivasoftware.com>
 *
 * See LICENSE file for license related information.
 *
 */

#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/prctl.h>
#include <libubox/uloop.h>
#include <libubus.h>
#include <libbbfdm-ubus/bbfdm-ubus.h>
#include <libbbfdm-api/dmcommon.h>

#include "config.h"
#include "utils.h"

extern DM_MAP_OBJ tDynamicObj[];

static struct bbfdm_context bbfdm_ctx = {0};
static void bulkdata_profile_cb(struct uloop_timeout *timeout);
static struct bulkdata bulkdata_main = {0};
static struct ubus_event_handler ev = {0};

LIST_HEAD(failedreports);

static int get_retry_period(int min)
{
	unsigned int random_value = 0;

	int fd = open("/dev/urandom", O_RDONLY);
	if (fd < 0) {
		return min;
	}

	if (read(fd, &random_value, sizeof(random_value)) != sizeof(random_value)) {
		close(fd);
		return min;
	}

	close(fd);

	// Generate a random number in the range [min, 2*min)
	return (random_value % min) + min;
}

static void bulkdata_run_profiles(struct bulkdata *bulkdata)
{
	if (bulkdata->enable == false) {
		bulkdata_set_status("Disabled");
		return;
	}

	bulkdata_set_status("Enabled");

	for (int i = 0; i < bulkdata->profile_number; i++) {
		bulkdata->profile[i].utimer.cb = bulkdata_profile_cb;
		
		bulkdata->profile[i].failed_reports = &failedreports;
		unsigned int next_period = get_next_period(bulkdata->profile[i].time_reference, bulkdata->profile[i].reporting_interval);
		INFO("The session of profile name '%s' will be start in %d sec", bulkdata->profile[i].profile_name, next_period);
		uloop_timeout_set(&bulkdata->profile[i].utimer, next_period * 1000);
	}
}

static void bulkdata_stop_profiles(struct bulkdata *bulkdata)
{
	for (int i = 0; i < bulkdata->profile_number; i++) {
		uloop_timeout_cancel(&bulkdata->profile[i].utimer);
		if (bulkdata->profile[i].c_process.pid != 0) {
			uloop_process_delete(&bulkdata->profile[i].c_process);
			kill(bulkdata->profile[i].c_process.pid, SIGTERM);
		}
	}
}

static void child_complete_cb(struct uloop_process *p, __attribute__((unused)) int ret)
{
	struct profile *profile = container_of(p, struct profile, c_process);
	if (profile == NULL)
		return;

	memset(&profile->c_process, 0, sizeof(struct uloop_process));
	uloop_timeout_set(&profile->utimer, 1000 * profile->next_session_delay);
	profile->next_session_delay = 0;
}

static void bulkdata_profile_cb(struct uloop_timeout *timeout)
{
	struct profile *profile = NULL;
	pid_t child;

	profile = container_of(timeout, struct profile, utimer);
	if (profile == NULL)
		return;

	child = fork();

	if (child == -1)
		return;

	if (child > 0) {
		// parent process
		memset(&profile->c_process, 0, sizeof(struct uloop_process));
		profile->c_process.pid = child;
		profile->c_process.cb = child_complete_cb;
		uloop_process_add(&profile->c_process);
		return;
	}

	// child process
	prctl(PR_SET_NAME, "bulkdata_helper");
	uloop_done();
	ubus_shutdown(&bbfdm_ctx.ubus_ctx);
	fclose(stderr);
	fclose(stdout);
	fclose(stdin);

	unsigned int http_code = 0;
	char *report = NULL;
	time_t now = time(NULL);

	INFO("New session of profile name '%s' started", profile->profile_name);
	if (profile->retry_count == 0 || profile->next_retry > now || !profile->http_retry_enable) //Perdiodic execution
		create_encoding_bulkdata_report(profile, &report);
	else
		create_failed_report(profile, &report);

	DEBUG("The content of the profile name report '%s' is :\n==========\n%s\n==========\n", profile->profile_name, report);
	http_code = http_send_message(profile, report);
	if (http_code != 200) {
		unsigned int retry_period = 0;

		if (profile->retry_count == 0 || profile->next_retry > now || !profile->http_retry_enable) {  //Perdiodic execution
			retry_period = get_retry_period(profile->http_retry_minimum_wait_interval);
			profile->next_period = now + profile->reporting_interval;
			profile->next_retry = now + retry_period;
			profile->retry_count = 1;
			profile->min_retry = profile->http_retry_minimum_wait_interval * 2;
			if ((profile->next_retry < profile->next_period) && profile->http_retry_enable) {
				INFO("Retry session of profile name '%s' in %d sec", profile->profile_name, retry_period);
				profile->next_session_delay = retry_period;
			} else {
				INFO("Start New session of profile name '%s' in %d sec", profile->profile_name, profile->reporting_interval);
				profile->next_session_delay = profile->reporting_interval;
			}
		} else { //Retry execution
			retry_period = get_retry_period(profile->min_retry);
			profile->min_retry*=2;
			profile->next_retry+=retry_period;
			profile->retry_count++;
			if (profile->next_retry < profile->next_period) {
				INFO("Retry session of profile name '%s' in %d sec", profile->profile_name, retry_period);
				profile->next_session_delay = retry_period;
			} else {
				INFO("Retry session of profile name '%s' in %ld sec", profile->profile_name, (profile->next_period-profile->next_retry+retry_period));
				profile->next_session_delay = profile->next_period - profile->next_retry + retry_period;
			}
		}

		if (profile->new_report) {
			bulkdata_add_failed_report(profile, profile->new_report);
			FREE(profile->new_report);
		}
		FREE(report);
	} else {
		if (profile->retry_count == 0 || profile->next_retry > now || !profile->http_retry_enable) {
			INFO("Start New session of profile name '%s' in %d sec", profile->profile_name, profile->reporting_interval);
			profile->next_session_delay = profile->reporting_interval;
		} else {
			INFO("Retry session of profile name '%s' in %ld sec", profile->profile_name, (profile->next_period-profile->next_retry));
			profile->next_session_delay = profile->next_period - profile->next_retry;
		}
		FREE(profile->new_report);
		FREE(report);
		empty_failed_reports_list(profile);
		profile->retry_count= 0;
	}
}

static void config_reload_cb(struct ubus_context *ctx __attribute__((unused)),
			struct ubus_event_handler *ev __attribute__((unused)),
			const char *type __attribute__((unused)),
			struct blob_attr *msg __attribute__((unused)))
{
	INFO("Reloading bulkdata upon bulkdata.reload event");
	bulkdata_stop_profiles(&bulkdata_main);
	bulkdata_config_fini(&bulkdata_main);
	bulkdata_config_init(&bulkdata_main);
	bulkdata_run_profiles(&bulkdata_main);
}

static int register_config_change(struct ubus_context *uctx)
{
	int ret;

	memset(&ev, 0, sizeof(struct ubus_event_handler));
	ev.cb = config_reload_cb;

	ret = ubus_register_event_handler(uctx, &ev, "bulkdata.reload");
	if (ret)
		return -1;

	return 0;
}

int main(void)
{
	int error = 0;
	bulkdata_set_status("Error");

	/* Logging to syslog */
	openlog("bulkdata", LOG_CONS | LOG_PID | LOG_NDELAY, LOG_LOCAL1);

	INFO("Start bulkdatad daemon");
	error = wait_for_bbfdm_object();
	if (error)
		return error;

	memset(&bbfdm_ctx, 0, sizeof(struct bbfdm_context));
	bbfdm_ubus_set_service_name(&bbfdm_ctx, "bulkdata");
	bbfdm_ubus_set_log_level(3);
	bbfdm_ubus_load_data_model(tDynamicObj);

	if (bbfdm_ubus_regiter_init(&bbfdm_ctx))
		goto out;

	if (register_config_change(&bbfdm_ctx.ubus_ctx) != 0)
		goto out;

	bulkdata_config_init(&bulkdata_main);

	bulkdata_run_profiles(&bulkdata_main);
	uloop_run();

out:
	bbfdm_ubus_regiter_free(&bbfdm_ctx);
	bulkdata_config_fini(&bulkdata_main);
	INFO("Stop bulkdatad daemon");
	closelog();
	return 0;
}
